//Copyright (C) 2016 Sinia Petri
#ifndef SP_GEOMETRY_H
#define SP_GEOMETRY_H
#include "ASMVars.h"
#include "VoronoiDiagramGenerator.h"
class spGeometry
  {
  public:
  struct TLine
	 {
	 int x1;
	 int y1;
	 int x2;
	 int y2;
	 };
  //---------------------------------------------------------------------------
  static void RotatePoints(TPoint *points, int points_number, TPoint *origin, double angle)
	{
	int ox, oy, x1, y1;
	ox = origin->x;
	oy = origin->y;
	for (int i = 0; i < points_number; i++)
		{
		x1 = points[i].x;
		y1 = points[i].y;
		points[i].x = fast_ftol(ox + (x1 - ox)*cos(angle) - (y1 - oy)*sin(angle));
		points[i].y = fast_ftol(oy + (x1 - ox)*sin(angle) + (y1 - oy)*cos(angle));
		}
	}
  //---------------------------------------------------------------------------
  static void GetRotateData(TPoint *point, TPoint *origin, double angle, TPoint *rotoData, bool relative)
	{
	int ox, oy, x1, y1;
	ox = origin->x;
	oy = origin->y;
	x1 = point->x;
	y1 = point->y;
	rotoData->x = fast_ftol(ox + (x1 - ox)*cos(angle) - (y1 - oy)*sin(angle));
	rotoData->y = fast_ftol(oy + (x1 - ox)*sin(angle) + (y1 - oy)*cos(angle));
	if (relative)
	   {
	   rotoData->x = rotoData->x - x1;
	   rotoData->y = rotoData->y - y1;
	   }
	}
  //---------------------------------------------------------------------------
  static void GetRotateData(TPoint *point, TPoint *origin, double angle, double *rotoX, double *rotoY, bool relative)
	{
	double ox, oy, x1, y1;
	ox = (double)origin->x;
	oy = (double)origin->y;
	x1 = (double)point->x;
	y1 = (double)point->y;
	*rotoX = ox + (x1 - ox)*cos(angle) - (y1 - oy)*sin(angle);
	*rotoY = oy + (x1 - ox)*sin(angle) + (y1 - oy)*cos(angle);
	if (relative)
	   {
	   *rotoX = *rotoX - x1;
	   *rotoY = *rotoY - y1;
	   }
	}
  //---------------------------------------------------------------------------
  static void TranslatePoints(TPoint *points, int points_number, TPoint *trans)
	{
	for (int i = 0; i < points_number; i++)
		{
		points[i].x += trans->x;
		points[i].y += trans->y;
		}
	}
  //---------------------------------------------------------------------------
  static void RotateOneTranslateAll(TPoint *points, int points_number, TPoint *origin, double angle)
	{
	int ox, oy, x1, y1;
	ox = origin->x;
	oy = origin->y;
	x1 = points[0].x;
	y1 = points[0].y;
	// rotate first
	points[0].x = fast_ftol(ox + (x1 - ox)*cos(angle) - (y1 - oy)*sin(angle));
	points[0].y = fast_ftol(oy + (x1 - ox)*sin(angle) + (y1 - oy)*cos(angle));
	//
	x1 = points[0].x - x1;
	y1 = points[0].y - y1;
	for (int i = 1; i < points_number; i++)
		{
		points[i].x += x1;
		points[i].y += y1;
		}
	}
  //---------------------------------------------------------------------------
  static void ReflectPoints(TPoint *points, int points_number, TPoint *origin, double angle)
	{
	int ox, oy, x1, y1;
	ox = origin->x;
	oy = origin->y;
	for (int i = 0; i < points_number; i++)
		{
		x1 = points[i].x;
		y1 = points[i].y;
		points[i].x = fast_ftol(ox + (x1 - ox)*cos(2.0*angle) + (y1 - oy)*sin(2.0*angle));
		points[i].y = fast_ftol(oy + (x1 - ox)*sin(2.0*angle) - (y1 - oy)*cos(2.0*angle));
		}
	}
  //---------------------------------------------------------------------------
  static void ChangeLineLength(TPoint *pointS, TPoint *pointE, int nLength)
	{
	int x = pointE->x - pointS->x, y = pointE->y - pointS->y;
	float hypo = sqrt(x*x + y*y);
	x = fast_ftol(1.0 * x * nLength/hypo);
	y = fast_ftol(1.0 * y * nLength/ hypo);
	pointE->x = x;
	pointE->y = y;
	}
  //---------------------------------------------------------------------------
  static void GetLineData(double *OutRadius, double *OutAngle, int x1, int y1, int x2, int y2)
	{
	//
	//  Izraunajmo x i y od linije i njenu hipotenuzu.
	//  Unda, izraunaj odma lipo kut
	//
	int x_ord, y_ord;
	double hypo, angle;
	y_ord = y1 - y2;
	x_ord = x2 - x1;
	hypo = sqrt(y_ord*y_ord + x_ord*x_ord);
	if (hypo == 0)
		{
		*OutAngle = 0;
		*OutRadius = 0;
		return;
		}
	// sve u prvom kvadrantu
	angle = asin((double)abs(y_ord) / hypo);
	//angle = 180.0*angle/M_PI; u stupnjevin
	//
	// sad izraunajmo koji je kvadrant
	//
	// prvi ili drugi kvadrant
	if (y_ord >= 0)
		{
		if (x_ord < 0)
		   angle = M_PI - angle;
			//angle = 180.0 - angle;
		}
	// trei ili etvrti
	else
		{
		angle = - angle;
		if (x_ord < 0)
			//angle = -180.0 - angle;
			angle = -M_PI - angle;
		}
	*OutAngle = angle;
	*OutRadius = hypo;
	//
	// A ta ako su koordinate linije ile priko mire ?
	//
	}
  //---------------------------------------------------------------------------
  static void VoronoiGenerate(std::vector<TPoint> &inpPoints, std::vector<TLine> &outLines, std::vector<Triangle> &outTriples,
							  int width, int height, float minDistance)
	{
	int points_number = inpPoints.size();
	float *xValues, *yValues;
	xValues = new float[points_number];
	yValues = new float[points_number];
	for (int i = 0; i < points_number; i++)
		{
		xValues[i] = inpPoints[i].x;
		yValues[i] = inpPoints[i].y;
		}
	VoronoiDiagramGenerator vdg;
	vdg.generateVoronoi(xValues, yValues, points_number, 0, width, 0, height, minDistance);
	delete []xValues;
	delete []yValues;
	vdg.resetIterator();
	TLine line;
	float x1, y1, x2, y2;
	while (vdg.getNext(x1, y1, x2, y2))
		{
		line.x1 = fast_ftol(x1);
		line.y1 = fast_ftol(y1);
		line.x2 = fast_ftol(x2);
		line.y2 = fast_ftol(y2);
		outLines.push_back(line);
		}
	// output Delaunay triangulation
	for (size_t i = 0; i < vdg.tripleList.size(); i++)
		outTriples.push_back(vdg.tripleList[i]);  // test
	}
  private:
	inline bool static yAscTPoint(TPoint lhs, TPoint rhs) { return lhs.y < rhs.y;}
  };
#endif
