//Copyright (C) 2016 Sinia Petri
#ifndef SP_PIXEL_ARITHMETICS_H
#define SP_PIXLE_ARITHMETICS_H
#include "spImage.hpp"
#include "spPixelTypes.hpp"
#include "spByteLogic.h"
#include <vector>
#include <algorithm>
#include "ASMVars.h"
class spPixelArithmetics
  {
  public:
	// structures
	template <typename T>
	struct spsFilter
	   {
	   // 3x3 filter sutructure
	   bool wrap;
	   bool chanOn[3];
	   float kernel[9];
	   float div;
	   float bias;
	   T nbh[9];
	   // logic filter
	   int  logic[9];
	   int  sign[9];
	   };
	template <typename T>
	struct spsArith
	   {
	   // standard arithmetics
	   T topVal;
	   T botVal;
	   bool invertPixel;
	   int bitShift;
	   // bumping paremeters
	   T bumpVal;
	   float bumpScale;
	   };
	template <typename T>
	struct spsColor
	   {
	   // color casting
	   T castVal[3];
	   bool castOn[3];
	   float castRange[2];
	   // color mapping
	   double mapFactor[3];
	   bool  mapRelative;
	   // erase match
	   T testVal;
	   int tolerance;
	   bool match;
	   // color scaling
	   float scaleRange[3][2];
	   };
	template <typename T>
	struct Execute
		{
		typedef void (*ArithFunc)(const T &, T &, void *);
		};
  //---------------------------------------------------------------------------
  // color normalization - standard
  //---------------------------------------------------------------------------
  Byte static NormColor(int colNew, Byte colBot, Byte colTop)
	{
	//
	// 	Funkcijica koja normalizira boju u granicama :
	//	Boja < colBot    => Boja = colBot
	//	Boja > colTop    => Boja = colTop
	//
	if (colTop <= colBot)
		return 0x00;
	if (colNew > ((int)colTop))
		return colTop;
	if (colNew < ((int)colBot))
		return colBot;
	return (Byte)colNew;
	}
  //---------------------------------------------------------------------------
  // color normalization - standard (template) - moda zatreba
  //---------------------------------------------------------------------------
  template <class T>
  void static NormColor(T &outVal, T &inpVal, T &botVal, T &topVal)
	{
	//
	// 	Funkcijica koja normalizira boju u granicama :
	//	Boja < colBot    => Boja = colBot
	//	Boja > colTop    => Boja = colTop
	//
	for (size_t k = 0; k < sizeof(T); k++)
		{
		if (topVal.ch[k] <= botVal.ch[k])
			outVal.ch[k] =  0;
		else if (inpVal.ch[k] > topVal.ch[k])
			outVal.ch[k] = topVal.ch[k];
		else if (inpVal.ch[k] < botVal.ch[k])
			outVal.ch[k] = botVal.ch[k];
		else
			outVal.ch[k] = inpVal.ch[k];
		}
	}
  //---------------------------------------------------------------------------
  // color normalization - circular
  //---------------------------------------------------------------------------
  Byte static CircColor(int colNew, Byte &colBot, Byte &colTop)
	{
	//
	//	Funkcijica koja normalizira boju u granicama :
	//	Boja < colBot   => oduzme razliku od colTop dok ne doe u granicu
	//	Boja > colTop   => dodaje razliku na colBot dok ne doe u granicu
	//
	if (colTop <= colBot)
		return 0x00;
	int interval, colLow, colHig;
	Byte colPix;
	interval = (int)colTop - (int)colBot;
	if (colNew > ((int)colTop))
		{
		colHig = colNew - (int)colTop;
		colNew = (int) colBot + (colHig % interval);
		}
	if (colNew < ((int)colBot))
		{
		colLow = (int)colBot - colNew;
		colNew = (int)colTop - (colLow % interval);
		}
	colPix = (BYTE)colNew;
	return colPix;
	}
  //---------------------------------------------------------------------------
  // color normalization - swap
  //---------------------------------------------------------------------------
  Byte static SwapColor(int colNew, Byte colBot, Byte colTop)
	{
	//
	//	Funkcijica koja normalizira boju u granicama :
	//	Boja < colBot	=> Boja = colTop
	//	Boja > colTop	=> Boja = colBot
	//	Dakle obratno od NormColor
	//
	if (colTop <= colBot)
		return 0x00;
	Byte colPix;
	if (colNew > ((int)colTop)) colPix = colBot;
	else if (colNew < ((int)colBot)) colPix = colTop;
	else colPix = (BYTE)colNew;
	return colPix;
	}
  //---------------------------------------------------------------------------
  // RGB to HSV conversion
  //---------------------------------------------------------------------------
  template <class T>
  static void Pixel_RGB_HSV(T &val, float *hsv)
	{
	float v, s, h, delta, r, g, b, rgb_max, rgb_min;
	r = (float)val.ch[2]  / 255.0;
	g = (float)val.ch[1]  / 255.0;
	b = (float)val.ch[0]  / 255.0;
	rgb_max = std::max(std::max(r, g), b);
	rgb_min = std::min(std::min(r, g), b);
	delta = rgb_max - rgb_min;
	v = rgb_max;
	hsv[2] = v;
	if ((delta > 0)&&(rgb_max >0))
		s = delta / rgb_max;
	else
	   {
	   s = 0;
	   h = -1;                    // undefined
	   hsv[0] = h;
	   hsv[1] = s;
	   return;
	   }
	if (r == rgb_max)
		h = (g - b)/delta;
	else if (g == rgb_max)
		h = 2 + (b - r)/delta;
	else
		h = 4 + ( r - g ) / delta;
	h = h*60;
	if (h < 0)
		h += 360;
	hsv[0] = h;
	hsv[1] = s;
	}
  //---------------------------------------------------------------------------
  // RGB to HSV conversion
  //---------------------------------------------------------------------------
  template <class T>
  static void Pixel_HSV_RGB(T &val, float *hsv)
	{
	int k, i;
	float f, p, q, t, r, g, b, h, s, v;
	h = hsv[0];
	s = hsv[1];
	v = hsv[2];
	if ( s == 0 && h == -1)  // undefined
	  {
	  for (k = 0; k < 3; k++)
		   val.ch[k] = (Byte)fast_ftol(v*255.0);
	  return;
	  }
	h /= 60;                        // sector 0 to 5
	i = floor( h );
	f = h - i;                      // factorial part of h
	p = v * ( 1 - s );
	q = v * ( 1 - s * f );
	t = v * ( 1 - s * ( 1 - f ) );
	switch( i )
		{
		case 0:
			r = v;
			g = t;
			b = p;
			break;
		case 1:
			r = q;
			g = v;
			b = p;
			break;
	   case 2:
			r = p;
			g = v;
			b = t;
			break;
	   case 3:
			r = p;
			g = q;
			b = v;
			break;
	   case 4:
			r = t;
			g = p;
			b = v;
			break;
	   default:
			r = v;
			g = p;
			b = q;
			break;
	   }
	val.ch[0] = (Byte)fast_ftol(b*255.0);
	val.ch[1] = (Byte)fast_ftol(g*255.0);
	val.ch[2] = (Byte)fast_ftol(r*255.0);
	}
  //---------------------------------------------------------------------------
  template <class T>
  static void Temperature_RGB(double temperature, T &val)
	{
	/*
	Algorithm: Tanner Helland, http://www.tannerhelland.com/4435/convert-temperature-rgb-algorithm-code/
	Start with a temperature, in Kelvin, somewhere between 1000 and 40000.  (Other values may work,
	but I can't make any promises about the quality of the algorithm's estimates above 40000 K.)
	Note also that the temperature and color variables need to be declared as floating-point.
	*/
	temperature = temperature / 100.0;
	double red, green, blue;
	// calculate red
	if (temperature <= 66)
		red = 255;
	else
		{
		red = temperature - 60.0;
		red = 329.698727446 * pow(red, -0.1332047592);
		if (red< 0)
		   red = 0;
		else if (red > 255)
		   red = 255;
		}
	// calculate green
	if (temperature <= 66)
	   {
	   green = temperature;
	   green = 99.4708025861 * log(green) - 161.1195681661;
	   if (green < 0)
		   green = 0;
	   else if (green > 255)
		   green = 255;
	   }
	else
	   {
	   green = temperature - 60;
	   green = 288.1221695283 * pow(green, -0.0755148492);
	   if (green < 0)
		  green = 0;
	   if (green > 255)
		  green = 255;
	   }
	// calculate blue
	if (temperature >= 66)
	   blue = 255;
	else
	   {
	   if (temperature <= 19)
		  blue = 0;
	   else
		  {
		  blue = temperature - 10;
		  blue = 138.5177312231 * log(blue) - 305.0447927307;
		  if (blue < 0)
			 blue = 0;
		  if (blue > 255)
			 blue = 255;
		  }
	   }
	val.ch[0] = (Byte)fast_ftol(blue);
	val.ch[1] = (Byte)fast_ftol(green);
	val.ch[2] = (Byte)fast_ftol(red);
	}
  //---------------------------------------------------------------------------
  // pixel arithmetics block
  //---------------------------------------------------------------------------
  template <class T>
  void static Pixel_NORM(const T &inpVal, T &outVal, void *data)
	{
	// NOP
	}
  //---------------------------------------------------------------------------
  template <class T>
  void static Pixel_SCALED(const T &inpVal, T &outVal,  void *data)
	{
	spsArith<T> *arith = static_cast <spsArith<T>*>(data);
	for (size_t k = 0; k < sizeof(T); k++)
		outVal.ch[k] = NormColor(outVal.ch[k], arith->botVal.ch[k], arith->topVal.ch[k]);
	}
  //---------------------------------------------------------------------------
  template <class T>
  void static Pixel_XOR(const T &inpVal, T &outVal, void *data)
	{
	spsArith<T> *arith = static_cast <spsArith<T>*>(data);
	for (size_t k = 0; k < sizeof(T); k++)
		outVal.ch[k] = NormColor(BitShiftInvert(inpVal.ch[k], arith) ^ (int)outVal.ch[k], arith->botVal.ch[k], arith->topVal.ch[k]);
	}
  //---------------------------------------------------------------------------
  template <class T>
  void static Pixel_AND(const T &inpVal, T &outVal, void *data)
	{
	spsArith<T> *arith = static_cast <spsArith<T>*>(data);
	for (size_t k = 0; k < sizeof(T); k++)
		outVal.ch[k] = NormColor(BitShiftInvert(inpVal.ch[k], arith) & (int)outVal.ch[k], arith->botVal.ch[k], arith->topVal.ch[k]);
	}
  //---------------------------------------------------------------------------
  template <class T>
  void static Pixel_OR(const T &inpVal, T &outVal, void *data)
	{
	spsArith<T> *arith = static_cast <spsArith<T>*>(data);
	for (size_t k = 0; k < sizeof(T); k++)
		outVal.ch[k] = NormColor(BitShiftInvert(inpVal.ch[k], arith) | (int)outVal.ch[k], arith->botVal.ch[k], arith->topVal.ch[k]);
	}
  //---------------------------------------------------------------------------
  template <class T>
  void static Pixel_ADD(const T &inpVal, T &outVal, void *data)
	{
	spsArith<T> *arith = static_cast <spsArith<T>*>(data);
	for (size_t k = 0; k < sizeof(T); k++)
		outVal.ch[k] = NormColor(BitShiftInvert(inpVal.ch[k], arith) + (int)outVal.ch[k], arith->botVal.ch[k], arith->topVal.ch[k]);
	}
  //---------------------------------------------------------------------------
  template <class T>
  void static Pixel_SUB(const T &inpVal, T &outVal, void *data)
	{
	spsArith<T> *arith = static_cast <spsArith<T>*>(data);
	for (size_t k = 0; k < sizeof(T); k++)
		outVal.ch[k] = NormColor(BitShiftInvert(inpVal.ch[k], arith) - (int)outVal.ch[k], arith->botVal.ch[k], arith->topVal.ch[k]);
	}
  //---------------------------------------------------------------------------
  template <class T>
  void static Pixel_MULT(const T &inpVal, T &outVal, void *data)
	{
	spsArith<T> *arith = static_cast <spsArith<T>*>(data);
	for (size_t k = 0; k < sizeof(T); k++)
		outVal.ch[k] = NormColor(BitShiftInvert(inpVal.ch[k], arith) * (int)outVal.ch[k], arith->botVal.ch[k], arith->topVal.ch[k]);
	}
  //---------------------------------------------------------------------------
  template <class T>
  void static Pixel_MULT_nA(const T &inpVal, T &outVal, void *data)
	{
	spsArith<T> *arith = static_cast <spsArith<T>*>(data);
	size_t  k;
	Byte max_value = 0x00;
	for (k = 0; k < sizeof(T); k++)
		{
		if (inpVal.ch[k] > max_value)
		   max_value = inpVal.ch[k];
		}
	for (k = 0, i = 3; k < 3; k++, i++)
		{
		if (max_value > 0x00)
		   outVal.ch[k] = NormColor(fast_ftol((float)BitShiftInvert(inpVal.ch[k], arith) * (float)outVal.ch[k] / (float) max_value), arith->botVal.ch[k], arith->topVal.ch[k]);
		}
	}
  //---------------------------------------------------------------------------
  template <class T>
  void static Pixel_MULT_nB(const T &inpVal, T &outVal, void *data)
	{
	spsArith<T> *arith = static_cast <spsArith<T>*>(data);
	size_t  k;
	Byte max_value = 0x00;
	for (k = 0; k < sizeof(T); k++)
		{
		if (outVal.ch[k] > max_value)
		   max_value = outVal.ch[k];
		}
	for (k = 0, i = 3; k < 3; k++, i++)
		{
		if (max_value > 0x00)
		   outVal.ch[k] = NormColor(fast_ftol((float)BitShiftInvert(inpVal.ch[k], arith) * (float)outVal.ch[k] / (float) max_value), arith->botVal.ch[k], arith->topVal.ch[k]);
		}
	}
  //---------------------------------------------------------------------------
  template <class T>
  void static Pixel_Higher(const T &inpVal, T &outVal, void *data)
	{
	spsArith<T> *arith = static_cast <spsArith<T>*>(data);
	for (size_t k = 0; k < sizeof(T); k++)
		{
		if (inpVal.ch[k] > outVal.ch[k])
		   outVal.ch[k] = NormColor(BitShiftInvert(inpVal.ch[k], arith), arith->botVal.ch[k], arith->topVal.ch[k]);
		}
	}
  //---------------------------------------------------------------------------
  template <class T>
  void static Pixel_Lower(const T &inpVal, T &outVal, void *data)
	{
	spsArith<T> *arith = static_cast <spsArith<T>*>(data);
	for (size_t k = 0; k < sizeof(T); k++)
		{
		if (inpVal.ch[k] < outVal.ch[k])
		   outVal.ch[k] = NormColor(BitShiftInvert(inpVal.ch[k], arith), arith->botVal.ch[k], arith->topVal.ch[k]);
		}
	}
  //---------------------------------------------------------------------------
  template <class T>
  void static Pixel_Mean(const T &inpVal, T &outVal, void *data)
	{
	spsArith<T> *arith = static_cast <spsArith<T>*>(data);
	for (size_t k = 0; k < sizeof(T); k++)
		outVal.ch[k] = NormColor(fast_ftol(((float)BitShiftInvert(inpVal.ch[k], arith) + (float)outVal.ch[k])/2.0), arith->botVal.ch[k], arith->topVal.ch[k]);
	}
  //---------------------------------------------------------------------------
  template <class T>
  void static Pixel_MinMax(const T &inpVal, T &outVal, void *data)
	{
	spsArith<T> *arith = static_cast <spsArith<T>*>(data);
	Byte b_max_value, e_max_value, b_min_value, e_min_value ;
	size_t k, e_max_elem, e_min_elem;
	bool gotMinMax;
	b_max_value = e_max_value = 0x00;
	b_min_value = e_min_value = 0xff;
	for (k = 0; k < sizeof(T); k++)
		{
		// maxs
		if (inpVal.ch[k] > b_max_value)
			b_max_value = inpVal.ch[k];
		if (outVal.ch[k] > e_max_value)
		   {
		   e_max_value = outVal.ch[k];
		   e_max_elem = k;
		   }
		// mins
		if (inpVal.ch[k] < b_min_value)
			b_min_value = pt_maper[k];
		if (outVal.ch[k] < e_min_value)
		   {
		   e_min_value = pt_bgr[k];
		   e_min_elem = k;
		   }
		}
	for (k = 0;  k < sizeof(T); k++)
		{
		gotMinMax = ((e_max_value < b_max_value) && (e_max_elem == k)) || ((e_min_value > b_min_value) && (e_min_elem == k));
		if (gotMinMax)
			outVal.ch[k] = NormColor(BitShiftInvert(inpVal.ch[k], arith), arith->botVal.ch[k], arith->topVal.ch[k]);
		}
	}
  //---------------------------------------------------------------------------
  template <class T>
  void static Pixel_Random(const T &inpVal, T &outVal, void *data)
	{
	spsArith<T> *arith = static_cast <spsArith<T>*>(data);
	if (random(2) == 0)
	   return;
	for (size_t k = 0; k < sizeof(T); k++)
		outVal.ch[k] = NormColor(BitShiftInvert(inpVal.ch[k], arith), arith->botVal.ch[k], arith->topVal.ch[k]);
	}
  //---------------------------------------------------------------------------
  template <class T>
  void static Pixel_Equalize(const T &inpVal, T &outVal, void *data)
	{
	spsArith<T> *arith = static_cast <spsArith<T>*>(data);
	float intensityA = 0.0, intensityB = 0.0;
	size_t k;
	float aByte;
	for (k = 0; k < sizeof(T); k++)
		{
		intensityA += (float)inpVal.ch[k];
		intensityB += (float)outVal.ch[k];
		}
	intensityA = intensityA/(3*255.0);
	intensityB = intensityB/(3*255.0);
	for (k = 0; k < sizeof(T); k++)
		{
		aByte = BitShiftInvert(inpVal.ch[k], arith) *intensityB;
		aByte += (float)outVal.ch[k] * intensityA;
		outVal.ch[k] = NormColor(fast_ftol(aByte),  arith->botVal.ch[k], arith->topVal.ch[k]);
		}
	}
  //---------------------------------------------------------------------------
  template <class T>
  void static Pixel_Burner(const T &inpVal, T &outVal, void *data)
	{
	spsArith<T> *arith = static_cast <spsArith<T>*>(data);
	float intensityA = 0.0, intensityB = 0.0;
	size_t k;
	for (k = 0; k < sizeof(T); k++)
		{
		intensityA += (float)inpVal.ch[k];
		intensityB += (float)outVal.ch[k];
		}
	if (intensityB > 0)
	   intensityB = intensityA/intensityB;
	for (k = 0; k < sizeof(T); k++)
		outVal.ch[k] = NormColor(fast_ftol((float)BitShiftInvert(inpVal.ch[k], arith) *intensityB), arith->botVal.ch[k], arith->topVal.ch[k]);
	}
  //---------------------------------------------------------------------------
  template <class T>
  void static Pixel_IntensityB(const T &inpVal, T &outVal, void *data)
	{
	spsArith<T> *arith = static_cast <spsArith<T>*>(data);
	float intensityB = 0;
	size_t k;
	for (k = 0; k < sizeof(T); k++)
		intensityB += (float)outVal.ch[k];
	intensityB = intensityB /(3*255.0);
	for (k = 0; k < sizeof(T); k++)
		outVal.ch[k] = NormColor(fast_ftol((float)BitShiftInvert(inpVal.ch[k], arith) *intensityB), arith->botVal.ch[k], arith->topVal.ch[k]);
	}
  //---------------------------------------------------------------------------
  template <class T>
  void static Pixel_IntensityA(const T &inpVal, T &outVal, void *data)
	{
	spsArith<T> *arith = static_cast <spsArith<T>*>(data);
	float intensityA = 0;
	size_t k;
	for (k = 0; k < sizeof(T); k++)
		intensityA += (float)BitShiftInvert(inpVal.ch[k], arith);
	intensityA = intensityA /(3*255.0);
	for (k = 0; k < sizeof(T); k++)
		outVal.ch[k] = NormColor(fast_ftol((float)outVal.ch[k] * intensityA), arith->botVal.ch[k], arith->topVal.ch[k]);
	}
  //---------------------------------------------------------------------------
  template <class T>
  void static Pixel_A_MOD_B(const T &inpVal, T &outVal, void *data)
	{
	spsArith<T> *arith = static_cast <spsArith<T>*>(data);
	for (size_t k = 0; k < sizeof(T); k++)
		{
		if (outVal.ch[k] > 0x00)
			outVal.ch[k] = NormColor(BitShiftInvert(inpVal.ch[k], arith) % (int)outVal.ch[k], arith->botVal.ch[k], arith->topVal.ch[k]);
		}
	}
  //---------------------------------------------------------------------------
  template <class T>
  void static Pixel_AXB(const T &inpVal, T &outVal, void *data)
	{
	spsArith<T> *arith = static_cast <spsArith<T>*>(data);
	int a_vect[3], b_vect[3], vect_vomp[3];
	size_t k;
	for (k = 0; k < sizeof(T); k++)
		{
		a_vect[k] = BitShiftInvert(inpVal.ch[k], arith);
		b_vect[k] = (int)outVal.ch[k];
		}
	vect_comp[2] = a_vect[1] * b_vect[0] - a_vect[0] *  b_vect[1];
	vect_comp[1] = a_vect[0] * b_vect[2] - a_vect[2] *  b_vect[0];
	vect_comp[0] = a_vect[2] * b_vect[1] - a_vect[1] *  b_vect[2];
	for (size_t k = 0; k < sizeof(T); k++)
		{
		if (outVal.ch[k] > 0x00)
			outVal.ch[k] = NormColor(vect_comp[k] > 0 ? vect_comp[k] : -vect_comp[k], arith->botVal.ch[k], arith->topVal.ch[k]);
		}
	}
  //---------------------------------------------------------------------------
  template <class T>
  void static Pixel_Bump(const T &inpVal, T &outVal, void *data)
	{
	spsArith<T> *arith = static_cast <spsArith<T>*>(data);
	if (arith->bumpScale < 0.001f)
	   return;
	float from_len = 0.0f, color_len = 0.0f;
	size_t k;
	for (k = 0; k < sizeof(T); k++)
		{
		from_len  = max(from_len,  (float)BitShiftInvert(inpVal.ch[k], arith));
		color_len = max(color_len, (float)arith->bumpVal.ch[k]); // shift color
		}
	from_len  *= arith->bumpScale;
	color_len *= arith->bumpScale;
	for (k = 0; k < sizeof(T); k++)
		outVal.ch[k] = NormColor(fast_ftol((float)outVal.ch[k] *(from_len - color_len)), arith->botVal.ch[k], arith->topVal.ch[k]);
	}
  //---------------------------------------------------------------------------
  template <class T>
  void static Pixel_BumpColor(const T &inpVal, T &outVal, void *data)
	{
	spsArith<T> *arith = static_cast <spsArith<T>*>(data);
	if (arith->bumpScale < 0.001f)
	   return;
	float from_len = 0.0f, color_len = 0.0f;
	size_t k;
	for (k = 0; k < sizeof(T); k++)
		{
		from_len  = max(from_len,  (float)BitShiftInvert(inpVal.ch[k], arith));
		color_len = max(color_len, (float)arith->bumpVal.ch[k]); // shift color
		}
	from_len  *= arith->bumpScale;
	color_len *= arith->bumpScale;
	for (k = 0; k < sizeof(T); k++)
		outVal.ch[k] = NormColor(fast_ftol((float)arith->bumpVal.ch[k] *(from_len - color_len)), arith->botVal.ch[k], arith->topVal.ch[k]);
	}
  //---------------------------------------------------------------------------
  template <class T>
  void static Pixel_BurnLinear(const T &inpVal, T &outVal, void *data)
	{
	spsArith<T> *arith = static_cast <spsArith<T>*>(data);
	for (size_t k = 0; k < sizeof(T); k++)
		outVal.ch[k] = NormColor((int)outVal.ch[k] + BitShiftInvert(inpVal.ch[k], arith) - 255,  arith->botVal.ch[k], arith->topVal.ch[k]);
	}
  //---------------------------------------------------------------------------
  template <class T>
  void static Pixel_SqrtAB(const T &inpVal, T &outVal, void *data)
	{
	spsArith<T> *arith = static_cast <spsArith<T>*>(data);
	float aByte;
	for (size_t k = 0; k < sizeof(T); k++)
		{
		aByte  =  (float)BitShiftInvert(inpVal.ch[k], arith);
		aByte *= aByte;
		aByte += (float)outVal.ch[k] * (float)outVal.ch[k];
		outVal.ch[k] = NormColor(fast_ftol(sqrt(aByte)),  arith->botVal.ch[k], arith->topVal.ch[k]);
		}
	}
  //---------------------------------------------------------------------------
  template <class T>
  void static Pixel_DodgeColor(const T &inpVal, T &outVal, void *data)
	{
	spsArith<T> *arith = static_cast <spsArith<T>*>(data);
	float aByte;
	for (size_t k = 0; k < sizeof(T); k++)
		{
		if (inpVal.ch[k] == 0xff)
		   continue;
		//aByte = (float)pt_maper[k] /(255.0 -  (float)pt_bgr[k]);
		aByte = (float)outVal.ch[k] /(255.0 -  (float)BitShiftInvert(inpVal.ch[k], arith));
		outVal.ch[k] = NormColor(fast_ftol(aByte),  arith->botVal.ch[k], arith->topVal.ch[k]);
		}
	}
  //---------------------------------------------------------------------------
  template <class T>
  void static Pixel_Screen(const T &inpVal, T &outVal, void *data)
	{
	spsArith<T> *arith = static_cast <spsArith<T>*>(data);
	float aByte;
	for (size_t k = 0; k < sizeof(T); k++)
		{
		aByte = 255.0 - (255.0 - (float)BitShiftInvert(inpVal.ch[k], arith)) * (255.0 -  (float)outVal.ch[k]);
		outVal.ch[k] = NormColor(fast_ftol(aByte),  arith->botVal.ch[k], arith->topVal.ch[k]);
		}
	}
  //---------------------------------------------------------------------------
  template <class T>
  void static Pixel_AbsSub(const T &inpVal, T &outVal, void *data)
	{
	spsArith<T> *arith = static_cast <spsArith<T>*>(data);
	for (size_t k = 0; k < sizeof(T); k++)
		outVal.ch[k] = NormColor(BitShiftInvert(abs(inpVal.ch[k], arith) - (int)outVal.ch[k], arith->botVal.ch[k]), arith->topVal.ch[k]);
	}
  //---------------------------------------------------------------------------
  template <class T>
  void static Pixel_Exsclusion(const T &inpVal, T &outVal, void *data)
	{
	spsArith<T> *arith = static_cast <spsArith<T>*>(data);
	float aByte;
	for (size_t k = 0; k < sizeof(T); k++)
		{
		if (inpVal.ch[k] == 0xff)
		   continue;
		aByte = 127.0 - 2.0*((float)BitShiftInvSrc(inpVal.ch[k], arith) - 127.0) * ((float)outVal.ch[k] - 127.0);
		outVal.ch[k] = NormColor(fast_ftol(aByte),  arith->botVal.ch[k], arith->topVal.ch[k]);
		}
	}
  //---------------------------------------------------------------------------
  template <class T>
  void static Pixel_DIV(const T &inpVal, T &outVal, void *data)
	{
	spsArith<T> *arith = static_cast <spsArith<T>*>(data);
	for (size_t k = 0; k < sizeof(T); k++)
		outVal.ch[k] = NormColor(fast_ftol((float)BitShiftInvert(inpVal.ch[k], arith) / (float)outVal.ch[k]), arith->botVal.ch[k], arith->topVal.ch[k]);
	}
  //---------------------------------------------------------------------------
  template <class T>
  void static Pixel_DIV_nA(const T &inpVal, T &outVal, void *data)
	{
	spsArith<T> *arith = static_cast <spsArith<T>*>(data);
	Byte max_value = 0x00;
	size_t k;
	for (k = 0; k < sizeof(T); k++)
		{
		if (inpVal.ch[k] > max_value)
			max_value = inpVal.ch[k];
		}
	for (k = 0; k < sizeof(T); k++)
		{
		if (outVal.ch[k] > 0x00)
		   outVal.ch[k] = NormColor(fast_ftol((double)BitShiftInvert(inpVal.ch[k], arith) * (double)max_value / (double)outVal.ch[k]), arith->botVal.ch[k], arith->topVal.ch[k]);
		}
	}
  //---------------------------------------------------------------------------
  // pixel color block
  //---------------------------------------------------------------------------
  template <class T>
  void static Pixel_Erase_MatchColor(const T &inpVal, T &outVal, void *data)
	{
	spsColor<T> *color = static_cast <spsColor<T>*>(data);
	int  pEdit, pTest;
	color->match = true;
	for (size_t k = 0; k < sizeof(T); k++)
		{
		pEdit = (int)inpVal.ch[k];
		pTest = (int)color->testVal.ch[k];
		color->match = color->match && (pEdit <= (pTest + tolerance) && pEdit >= (pTest - tolerance));
		}
	if (!color->match)
	   outVal = inpVal;
	}
  //---------------------------------------------------------------------------
  template <class T>
  void static Pixel_Color_MaxRG(const T &inpVal, T &outVal, void *data)
	{
	outVal = inpVal;
	if (outVal.ch[2] > outVal.ch[1])
		outVal.ch[1] = outVal.xh[2];
	else
		outVal.ch[2] = outVal.ch[1];
	}
  //---------------------------------------------------------------------------
  template <class T>
  void static Pixel_Color_MaxGB(const T &inpVal, T &outVal, void *data)
	{
	outVal = inpVal;
	if (outVal.ch[1] > outVal.ch[0])
		outVal.ch[0] = outVal.xh[1];
	else
		outVal.ch[1] = outVal.ch[0];
	}
  //---------------------------------------------------------------------------
  template <class T>
  void static Pixel_Color_MaxBR(const T &inpVal, T &outVal, void *data)
	{
	outVal = inpVal;
	if (outVal.ch[2] > outVal.ch[0])
		outVal.ch[0] = outVal.xh[2];
	else
		outVal.ch[2] = outVal.ch[0];
	}
  //---------------------------------------------------------------------------
  template <class T>
  void static Pixel_Color_MaxRGB(const T &inpVal, T &outVal, void *data)
	{
	Byte max_byte = 0x00;
	size_t k;
	for (k = 0; k < sizeof(T); k++)
		{
		if (inpVal.ch[k]) > max_byte)
		   max_byte = inpVal.ch[k];
		}
	for (k = 0; k < sizeof(T); k++)
		outVal.ch[k] = max_byte;
	}
  //---------------------------------------------------------------------------
  template <class T>
  void static Pixel_Color_MinRG(const T &inpVal, T &outVal, void *data)
	{
	outVal = inpVal;
	if (outVal.ch[2] < outVal.ch[1])
		outVal.ch[1] = outVal.xh[2];
	else
		outVal.ch[2] = outVal.ch[1];
	}
  //---------------------------------------------------------------------------
  template <class T>
  void static Pixel_Color_MinGB(const T &inpVal, T &outVal, void *data)
	{
	outVal = inpVal;
	if (outVal.ch[1] < outVal.ch[0])
		outVal.ch[0] = outVal.xh[1];
	else
		outVal.ch[1] = outVal.ch[0];
	}
  //---------------------------------------------------------------------------
  template <class T>
  void static Pixel_Color_MinBR(const T &inpVal, T &outVal, void *data)
	{
	outVal = inpVal;
	if (outVal.ch[2] < outVal.ch[0])
		outVal.ch[0] = outVal.xh[2];
	else
		outVal.ch[2] = outVal.ch[0];
	}
  //---------------------------------------------------------------------------
  template <class T>
  void static Pixel_Color_MinRGB(const T &inpVal, T &outVal, void *data)
	{
	Byte min_byte = 0xff;
	size_t k;
	for (k = 0; k < sizeof(T); k++)
		{
		if (inpVal.ch[k]) < min_byte)
		   min_byte = inpVal.ch[k];
		}
	for (k = 0; k < sizeof(T); k++)
		outVal.ch[k] = min_byte;
	}
  //---------------------------------------------------------------------------
  template <class T>
  void static Pixel_Color_ExcRG(const T &inpVal, T &outVal, void *data)
	{
	outVal = inpVal;
	outVal.ch[1] = outVal.ch[1]^outVal.ch[2];
	outVal.ch[2] = outVal.ch[1]^outVal.ch[2];
	outVal.ch[1] = outVal.ch[1]^outVal.ch[2];
	}
  //---------------------------------------------------------------------------
  template <class T>
  void static Pixel_Color_ExcGB(const T &inpVal, T &outVal, void *data)
	{
	outVal = inpVal;
	outVal.ch[0] = outVal.ch[0]^outVal.ch[1];
	outVal.ch[1] = outVal.ch[0]^outVal.ch[1];
	outVal.ch[0] = outVal.ch[0]^outVal.ch[1];
	}
  //---------------------------------------------------------------------------
  template <class T>
  void static Pixel_Color_ExcBR(const T &inpVal, T &outVal, void *data)
	{
	outVal = inpVal;
	outVal.ch[0] = outVal.ch[0]^outVal.ch[2];
	outVal.ch[2] = outVal.ch[0]^outVal.ch[2];
	outVal.ch[0] = outVal.ch[0]^outVal.ch[2];
	}
  //---------------------------------------------------------------------------
  template <class T>
  void static Pixel_Color_ExcRGB(const T &inpVal, T &outVal, void *data)
	{
	outVal = inpVal;
	Byte temp_byte = outVal.ch[2];
	outVal.ch[2] = outVal.ch[1];
	outVal.ch[1] = outVal.ch[0];
	outVal.ch[0] = temp_byte;
	}
  //---------------------------------------------------------------------------
  template <class T>
  void static Pixel_Color_ExcMinMax(const T &inpVal, T &outVal, void *data)
	{
	T tempVal;
	Byte max_byte = 0x00, min_byte = 0xff;
	int  max_elem = 0, min_elem = 0;
	size_t k;
	for (k = 0; k < sizeof(T); k ++)
		{
		tempVal.ch[k] = inpVal.ch[k];
		if (tempVal.ch[k] > max_byte)
		   {
			max_byte = tempVal.ch[k];
			max_elem = k;
		   }
		if (tempVal.ch[k] < min_byte)
		   {
			min_byte = tempVal.ch[k];
			min_elem = k;
		   }
		}
	tempVal.ch[max_elem] = tempVal.ch[max_elem]^tempVal.ch[min_elem];
	tempVal.ch[min_elem] = tempVal.ch[max_elem]^tempVal.ch[min_elem];
	tempVal.ch[max_elem] = tempVal.ch[max_elem]^tempVal.ch[min_elem];
	for (k = 0; k < sizeof(T); k++)
		outVal.ch[k] = tempVal.ch[k];
	}
  //---------------------------------------------------------------------------
  template <class T>
  void static Pixel_Color_Brightness(const T &inpVal, T &outVal, void *data)
	{
	spsColor<T> *color = static_cast <spsColor<T>*>(data);
	double adjust;
	/* iz stare pixopedie - provjeri
		factor = ((float)pt_bgr[k] - 100.0) *px_SP_NORM /100.0;
	if (pt_bgr[3])
	   adjust = (adjust << px_SP_SHIFT) + fast_ftol(adjust * factor);
	else
	   adjust = (adjust << px_SP_SHIFT) +  fast_ftol(255.0 * factor);
	adjust = adjust >> px_SP_SHIFT;
	*/
	for (size k = 0; k < sizeof(T); k++)
		{
		adjust = (double)inpVal.ch[k];
		if (color->mapRelative)
		   adjust = adjust + adjust * color->mapFactor[k];
		else
		   adjust = adjust + 255.0 * color->mapFactor[k];
		outVal.ch[k] = NormColor(fast_ftol(adjust), 0x00, 0xff);
		}
	}
  //---------------------------------------------------------------------------
  template <class T>
  void static Pixel_Color_Stretch(const T &inpVal, T &outVal, void *data)
	{
	spsColor<T> *color = static_cast <spsColor<T>*>(data);
	double adjust;
	for (size k = 0; k < sizeof(T); k++)
		{
		adjust = (double)inpVal.ch[k];
		if (color->mapRelative)
		   adjust = adjust + adjust * color->mapFactor[k] - (255.0 * color->mapFactor[k]/2.0);
		else
		   adjust = adjust + 255.0 * color->mapFactor[k] - (255.0 * color->mapFactor[k]/2.0);
		outVal.ch[k] = NormColor(fast_ftol(adjust), 0x00, 0xff);
		}
	}
  //---------------------------------------------------------------------------
  template <class T>
  void static Pixel_ColorLMH(const T &inpVal, T &outVal, void *data)
	{
	spsColor<T> *color = static_cast <spsColor<T>*>(data);
	double factor, dbl_bgr[3], max_elem = 0.0;
	size_t  k, index;
	for (k = 0; k < sizeof(T); k++)
		{
		dbl_bgr[k] = (double)inpVal.ch[k];
		if (dbl_bgr[k] > max_elem)
		   max_elem = dbl_bgr[k];
		}
	if (max_elem < 85)
		index = 2;
	else if ((max_elem >= 85) && (max_elem < 170))
		index = 1;
	else
		index = 0;
	for (k = 0; k < sizeof(T); k++)
		{
		if (color->mapRelative)
		   dbl_bgr[k] = dbl_bgr[k] + dbl_bgr[k] * color->mapFactor[index];
		else
		   dbl_bgr[k] = dbl_bgr[k] + 255.0 * color->mapFactor[index];
		outVal.ch[k] = NormColor(fast_ftol(dbl_bgr[k]), 0x00, 0xff);
		}
	}
  //---------------------------------------------------------------------------
  template <class T>
  void static Pixel_ColorHSV(const T &inpVal, T &outVal, void *data)
	{
	spsColor<T> *color = static_cast <spsColor<T>*>(data);
	float hsv[3];
	size_t  k;
	Pixel_RGB_HSV(inpVal, hsv);
	if (hsv[0] != -1 && hsv[1] != 0)
	   {
	   for (size_t k = 0; k < sizeof(T); k++)
		   {
		   if (color->mapRelative)
			  hsv[k] += (float)(hsv[k] * color->mapFactor[k]);
		   else
			  {
			  if (k == 0)
				 hsv[k] += (float)(color->mapFactor[k] * 360.0);
			  else
				 hsv[k] += (float)color->mapFactor[k];
			  }
		   hsv[k] = hsv[k] < 0.0 ? 0 : hsv[k];
		   if (k == 0)
			  hsv[k] = hsv[k] > 360.0 ? 360.0 : hsv[k];
		   else
			  hsv[k] = hsv[k] > 1 ? 1 : hsv[k];
		   }
	   }
	Pixel_HSV_RGB(outVal, hsv);
	}
  //---------------------------------------------------------------------------
  template <class T>
  void static Pixel_Gamma(const T &inpVal, T &outVal, void *data)
	{
	spsColor<T> *color = static_cast <spsColor<T>*>(data);
	size_t k;
	double fl_maper, fl_color;
	for (int k = 0; k < 3; k++)
		{
		fl_maper = (double)inpVal.ch[k]/255.0;
		fl_color = pow( fl_maper, 1.0/ color->factor[k])*255.0);
		pt_bgr[k] = NormColor(fast_ftol(fl_color), 0x00, 0xff);
		}
	}
  //---------------------------------------------------------------------------
  template <class T>
  void static Pixel_Cast(const T &inpVal, T &outVal, void *data)
	{
	spsColor<T> *color = static_cast <spsColor<T>*>(data);
	float intensity = 0.0f;
	// izraunaj intenzitet ulaznog pixela
	for (size_t k = 0; k < sizeof(T); k++)
		intensity += (float)inpVal.ch[k];
	intensity /= 3.0;
	// napravi cast
	if (color->castOn[0] && intensity <= color->castRange[0])
	   outVal = color->castVal[0];
	else if (color->castOn[1] && intensity > color->castRange[0] && intensity <= color->castRange[1])
	   outVal = color->castVal[1];
	else if (color->castOn[2] && intensity > color->castRange[1])
	   outVal = color->castVal[2];
	else
	   outVal = inpVal;
	}
  //---------------------------------------------------------------------------
  template <class T>
  void static Pixel_Scale(const T &inpVal, T &outVal, void *data)
	{
	spsColor<T> *color = static_cast <spsColor<T>*>(data);
	float scaledChan;
	for (size_t k = 0; k < sizeof(T); k++)
		{
		scaledChan = (float)inpVal.ch[k] * (scaleRange[k][1] - scaleRange[k][0])/255.0 + scaleRange[k][0];
		outVal.ch[k] = NormColor(fast_ftol(scaledChan), 0x00, 0xff);
		}
	}
  //---------------------------------------------------------------------------
  template <class T>
  void static Pixel_ColorLuminance(const T &inpVal, T &outVal, void *data)
	{
	spsColor<T> *color = static_cast <spsColor<T>*>(data);
	// koristimo castVal kao input za boje
	float luminaLeft = 0.0f, luminaRight = 0.0f, adjChan;
	size_t k;
	for (k = 0; k < sizeof(T); k++)
		{
		luminaLeft  += (float)inpVal.ch[k] * LUMINANCE[k];
		luminaRight += (float)outVal.ch[k] * LUMINANCE[k];
		}
	if (luminaRight <= 0.0f)
	   return;
	for (k = 0; k < sizeof(T); k++)
		{
		adjChan = (float)outVal.ch[k] * luminaLeft / luminaRight;
		outVal.ch[k] = NormColor(fast_ftol(adjChan), 0x00, 0xff);
		}
	}
  //---------------------------------------------------------------------------
  // linear filter 3x3 (convolution)
  //---------------------------------------------------------------------------
  template <class T>
  void static Filter_Linear33(const T &inpVal, T &outVal, void *data)
	{
	spsFilter<T> *filter = static_cast <spsFilter<T>*>(data);
	int m;
	float fPixel;
	for (size_t k = 0; k < sizeof(T); k++)
		{
		if (filter->chanOn[k])
		   {
		   fPixel = 0.0f;
		   for (m = 0; m < 9; m++)
			   fPixel += (float)filter->nbh[m].ch[k] * filter->kernel[m];
		   fPixel = fPixel / filter->div + filter->bias;
		   outVal.ch[k] = NormColor(fast_ftol(fPixel), 0x00, 0xff);
		   }
		else
		   outVal.ch[k] = inpVal.ch[k]; //(Byte)filter->nbh[4].ch[k];
	   }
	}
  //---------------------------------------------------------------------------
  // closest filter 3x3
  //---------------------------------------------------------------------------
  template <class T>
  void static Filter_Closest33(const T &inpVal, T &outVal, void *data)
	{
	spsFilter<T> *filter = static_cast <spsFilter<T>*>(data);
	int m, index;
	float fPixel, epsilon, epsilon_test;
	for (size_t k = 0; k < sizeof(T); k++)
		{
		if (filter->chanOn[k])
		   {
		   fPixel = 0.0f;	 // konvolucija
		   for (m = 0; m < 9; m++)
			   fPixel += (float)filter->nbh[m].ch[k] * filter->kernel[m];
		   fPixel = fPixel / filter->div;
		   // kanal najblii konvoluciji
		   epsilon = 9999.0f;
		   index = 0;
		   for (m = 0; m < 9; m++)
			   {
			   epsilon_test = fabs(filter->nbh[m].ch[k] - fabs(fPixel));
			   if (epsilon_test  < epsilon )
				   {
				   epsilon = epsilon_test;
				   index  = m;
				   }
				}
		   fPixel = filter->nbh[index].ch[k] + filter->bias;
		   outVal.ch[k] = NormColor(fast_ftol(fPixel), 0x00, 0xff);
		   }
		else
		   outVal.ch[k] = inpVal.ch[k]; //(Byte)filter->nbh[4].ch[k];
	   }
	}
  //---------------------------------------------------------------------------
  // closest filter 3x3
  //---------------------------------------------------------------------------
  template <class T>
  void static Filter_Diff33(const T &inpVal, T &outVal, void *data)
	{
	spsFilter<T> *filter = static_cast <spsFilter<T>*>(data);
	int m, index;
	float fPixel, mid_convo, epsilon_test;
	for (size_t k = 0; k < sizeof(T); k++)
		{
		if (filter->chanOn[k])
		   {
		   fPixel = 0.0f;	 // konvolucija
		   mid_convo = filter->nbh[4].ch[k] * filter->kernel[4];
		   for (m = 0; m < 9; m++)
			   {
			   if (m != 4)
				  {
				  epsilon_test = filter->nbh[m].ch[k] * filter->kernel[m];
				  if (fabs(fabs(epsilon_test) - fabs(mid_convo))  > fabs(fPixel))
					 fPixel = mid_convo - epsilon_test;
				  }
			   }
		   fPixel = fPixel / filter->div + filter->bias;;
		   outVal.ch[k] = NormColor(fast_ftol(fPixel), 0x00, 0xff);
		   }
		else
		   outVal.ch[k] = inpVal.ch[k]; //(Byte)filter->nbh[4].ch[k];
	   }
	}
  //---------------------------------------------------------------------------
  // closest filter 3x3
  //---------------------------------------------------------------------------
  template <class T>
  void static Filter_MinMax33(const T &inpVal, T &outVal, void *data)
	{
	spsFilter<T> *filter = static_cast <spsFilter<T>*>(data);
	int m;
	float fPixel, convo[9], min_value, max_value, divisor, min_fact, max_fact;
	for (size_t k = 0; k < sizeof(T); k++)
		{
		if (filter->chanOn[k])
		   {
		   divisor = 0;
		   convo[k] = 0.0f;
		   min_value = 99999.0f;
		   max_value = -99999.0f;
		   for (m = 0; m < 9; m++)
			   {
			   convo[k] += filter->nbh[m].ch[k] * filter->kernel[m];
			   divisor  += filter->kernel[m];
			   if (filter->nbh[m].ch[k] < min_value)
				  {
				  min_fact  = filter->kernel[m];
				  min_value = filter->nbh[m].ch[k];
				  }
			   if (filter->nbh[m].ch[k] > max_value)
				  {
				  max_fact  = filter->kernel[m];
				  max_value = filter->nbh[m].ch[k];
				  }
			   }
		   convo[k] = convo[k] -  max_value*max_fact - min_value*min_fact;
		   divisor = divisor - max_fact - min_fact;
		   convo[k] =  (fast_ftol(divisor)  == 0 ? convo[k] :  convo[k] / divisor);
		   fPixel =  (max_value - convo[k]) > (convo[k] - min_value) ? max_value : min_value;
		   fPixel = fPixel/filter->div + filter->bias;
		   outVal.ch[k] = NormColor(fast_ftol(fPixel), 0x00, 0xff);
		   }
		else
		   outVal.ch[k] = inpVal.ch[k]; //(Byte)filter->nbh[4].ch[k];
	   }
	}
  //---------------------------------------------------------------------------
  // closest filter 3x3
  //---------------------------------------------------------------------------
  template <class T>
  void static Filter_Logic33(const T &inpVal, T &outVal, void *data)
	{
	spsFilter<T> *filter = static_cast <spsFilter<T>*>(data);
	int m;
	float fPixel;
	Byte ker_result;
	for (size_t k = 0; k < sizeof(T); k++)
		{
		if (filter->chanOn[k])
		   {
		   fPixel = 0.0f;	 // konvolucija
		   for (m = 0; m < 9; m++)
			   {
			   ker_result = (Byte)fast_ftol(filter->kernel[m]);
			   ker_result = bLogic[filter->logic[m]] (filter->nbh[m].ch[k], ker_result);
			   fPixel += (float)ker_result * (float)filter->sign[m];
		   }
		   fPixel = fPixel / filter->div + filter->bias;;
		   outVal.ch[k] = NormColor(fast_ftol(fPixel), 0x00, 0xff);
		   }
		else
		   outVal.ch[k] = inpVal.ch[k]; //(Byte)filter->nbh[4].ch[k];
	   }
	}
  //TODO: kreni na brushbasic / surface i nakon toga idemo miki konvertiraj pix
  private:
	static float LUMINANCE[3];
  //---------------------------------------------------------------------------
  // shift/invert pixel
  //---------------------------------------------------------------------------
  template <class T>
  int static BitShiftInvert(Byte inPix, spsArith<T> *arith)
	{
	int outPix = (int)inPix;
	if (arith->invertPixel)
	   outPix = 255 - outPix;
	if (arith->bitShift == 0)
	   return outPix;
	arith->bitShift > 0 ? outPix = outPix<<arith->bitShift : outPix = outPix>>abs(arith->bitShift);
	return outPix;
	}
  };
float spPixelArithmetics::LUMINANCE[3] = {0.11, 0.59, 0.30};
#endif

